package net.sf.freetime;

import java.util.Vector;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;

/**
 * Similar to the java properties class but allows storage/reading
 * of CSV files of variable length records. Hey, this is what
 * PersonalTime uses so it's easier to go with it - we don't know
 * what all the values are so this way we don't throw anything
 * that might be needed in future.
 * 
 * Allows grouped searching and other cool things.
 * 
 * @author Robin Rawson-Tetley
 *
 */
public class CSVMap {

	/** The rows of CSV data in the map */
	private Vector rows = new Vector();
	
	public CSVMap() {
	}
	
	/**
	 * Loads it in from 
	 * @param f
	 */
	public CSVMap(File f) {
		loadFromFile(f);
	}
	
	/** Searchs the rows in the map for the specified
	 *  value in the specified column. Returns the
	 *  CSVRow or null if no match was found.
	 *  
	 * @param column The column to search in
	 * @param value The value to find
	 * @return The matching CSVRow or null if no match
	 */
	public CSVRow findFirst(int column, String value) {
		CSVRow c = null;
		for (int i = 0; i < rows.size(); i++) {
			c = (CSVRow) rows.get(i);
			if (c.getValue(column).equalsIgnoreCase(value)) {
				return c;
			}
		}
		return null;
	}
	
	/**
	 * Given a column number and value, returns
	 * all rows of that value 
	 * 
	 * @param value1
	 * @param value2
	 * @return
	 */
	public CSVRow[] find(int column , String value) {
		Vector r = new Vector();
		CSVRow c = null;
		for (int i = 0; i < rows.size(); i++) {
			c = (CSVRow) rows.get(i);
			if (c.getValue(column).equalsIgnoreCase(value)) {
				r.add(c);
			}
		}
		CSVRow[] v = new CSVRow[r.size()];
		for (int i = 0; i < r.size(); i++)
			v[i] = (CSVRow) r.get(i);
		return v;
	}
	
	/**
	 * Finds something in an array of CSV rows 
	 * @param findin
	 * @param column
	 * @param value
	 * @return
	 */
	public static CSVRow[] find(CSVRow[] findin, int column, String value) {
		Vector r = new Vector();
		CSVRow c = null;
		for (int i = 0; i < findin.length; i++) {
			c = (CSVRow) findin[i];
			if (c.getValue(column).equalsIgnoreCase(value)) {
				r.add(c);
			}
		}
		CSVRow[] v = new CSVRow[r.size()];
		for (int i = 0; i < r.size(); i++)
			v[i] = (CSVRow) r.get(i);
		return v;
	}
	
	/**
	 * Reads a CSV file to generate the map
	 * @param f
	 */
	public void loadFromFile(File f) {
		try {
			// Bail if we don't find the file - we'll
			// create it next time.
			if (!f.exists()) return;
			BufferedReader in = new BufferedReader(new FileReader(f));
			String r = null;
			do {
				r = in.readLine();
				if (r != null)
					rows.addElement(new CSVRow(r));
			} while (r != null);
		}
		catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Outputs this map to a file.
	 * @param f
	 */
	public void saveToFile(File f) {
		try {
			FileOutputStream out = new FileOutputStream(f);
			out.write(this.toString().getBytes());
			out.flush();
			out.close();
		}
		catch (Exception e) {
			e.printStackTrace();
		}	
	}
	
	/**
	 * Returns the row
	 * @param row
	 * @return
	 */
	public CSVRow get(int row) {
		return (CSVRow) rows.get(row);
	}
	
	/** Returns the number of rows */
	public int getCount() {
		return rows.size();
	}
	
	/**
	 * Adds a new row. If a row with the
	 * same data exists, it is ignored. 
	 * @param row
	 */
	public void append(CSVRow row) {
		CSVRow c = null;
		for (int i = 0; i < rows.size(); i++) {
			c = (CSVRow) rows.get(i);
			if (c.equals(row)) return;
		}
		rows.addElement(row);
	}
	
	/**
	 * Deletes the row at a specified position
	 * @param row
	 */
	public void remove(int row) {
		rows.removeElementAt(row);
	}
	
	/** Removes the given row */
	public void remove(CSVRow row) {
		rows.removeElement(row);
	}
	
	/**
	 * Produces a CSV file from the data
	 */
	public String toString() {
		StringBuffer b = new StringBuffer();
		for (int i = 0; i < rows.size(); i++) {
			if (i > 0) b.append("\r\n");
			b.append( ((CSVRow) rows.get(i)).toString() );
		}
		return b.toString();
	}

}
