/*
 * Created on Aug 30, 2005
 *
 * To change the template for this generated file go to
 * Window - Preferences - Java - Code Generation - Code and Comments
 */
package net.sf.freetime;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Properties;
import java.util.Vector;

import net.sf.freetime.dto.Expense;
import net.sf.freetime.dto.ExpenseType;
import net.sf.freetime.dto.Phase;
import net.sf.freetime.dto.Project;
import net.sf.freetime.dto.Task;
import net.sf.freetime.dto.Timesheet;
import net.sf.freetime.dto.VAT;
import net.sf.ooweb.util.Logger;

/**
 * Handles data operations for the FreeTime UI.
 * 
 * @author Robin Rawson-Tetley
 */
public class DataController {

	protected CSVMap data = new CSVMap();
	protected Logger log = null;
	
	/** The filename of the main datastore */
	private final static String DATAFILE_NAME = System.getProperty("user.home") + File.separator + ".freetime";
	/** The date format to use */
	private final static String DATE_FORMAT = "dd/MM/yyyy";
	/** The timewatch shared directory */
	private String sharedDirectory = null;
	/** The timewatch group name */
	private String group = null;
	
	/** Employee code */
	private String ecode = null;
	
	/** Employee name */
	private String ename = null;
	private String efirstname = null;
	private String esurname = null;
	
	/** The special code timewatch stamps on all communications */
	private String ptcode = null;
	
	/** Sequence of next timesheet batch post to stamp */
	private int nextSequence = 1;

	/**
	 * Loads in the data from the default
	 * datafile location and reads the rc file as well.
	 */
	public DataController(Logger log) {

		this.log = log;
		
		// Load the main database in from a file
		// called .freetime in the user's home
		// directory.
		log.info("Loading data from " + DATAFILE_NAME);
		data.loadFromFile(new File(DATAFILE_NAME));
		log.info("Found " + data.getCount() + " records.");
		
		parseEmployee();
		
		Properties p = null;
		try {
			p = new Properties();
			p.load(new FileInputStream(new File(System.getProperty("user.home") + File.separator + ".freetimerc")));
		}
		catch (Exception e) {
			System.err.print("Could not find $HOME/.freetimerc");
			System.exit(1);
		}
		
		try {
			sharedDirectory = p.getProperty("timewatch.directory");
			if (sharedDirectory.endsWith(File.separator))
				sharedDirectory = sharedDirectory.substring(0, sharedDirectory.length() -1);
			group = p.getProperty("timewatch.group");
			log.info("Read timewatch settings.");
		}
		catch (Exception e) {
			log.error(e);
		}
	}

	/**
	 * Reads the employee bits from the data and populates
	 * local variables for easy reading. This should be called
	 * after data updates to refresh the data.
	 */
	protected void parseEmployee() {
		// Get the employee's name and code
		CSVRow[] c = data.find(0, "E");
		ecode = "ERROR_NO_CODE_FOUND";
		ename = "ERROR_NO_NAME_FOUND";
		if (c.length > 0) {
			ecode = c[0].getValue(1);
			ename = (c[0].getValue(2) + " " + c[0].getValue(3));
			efirstname = c[0].getValue(2);
			esurname = c[0].getValue(3);
		}
		// Get the special timewatch code
		c = data.find(0, "AA");
		ptcode = "ERROR_PTCODE_NOT_FOUND";
		if (c.length > 0) {
			ptcode = c[0].getValue(1);
		}
		// Get the next sequence - this is one
		// we add and is not normally in a pts
		c= data.find(0, "SEQ");
		if (c.length > 0) {
			nextSequence = Integer.parseInt(c[0].getValue(1));
			log.info("Read next sequence from data: " + nextSequence);
		}
		else {
			CSVRow r = new CSVRow(2);
			r.putValue(0, "SEQ");
			r.putValue(1, "1");
			data.append(r);
			nextSequence = 1;
			data.saveToFile(new File(DATAFILE_NAME));
			log.info("No sequence found in data file, creating one");
		}
	}
	
	/**
	 * Reads the shared directory where update files can be 
	 * found and processes any. Returns the number of updates done.
	 * @return The number of updates found.
	 */
	public int updateFromShare() {
		// Look for an update file with the
		// group name in the shared directory
		String uf = sharedDirectory + File.separator + group + ".pts";
		log.info("Searching for updates in: " + uf);
		File f = new File(uf);
		if (!f.exists()) {
			log.info("No updates were found");
			return 0;
		}
		
		// Copy values from it into the main data store
		CSVMap m = new CSVMap(f);
		int i = 0;
		for (i = 0; i < m.getCount(); i++) {
			data.append(m.get(i));
			log.info("Merged line: " + m.get(i).toString());
		}
		
		// Rename the update file to .pbs
		String newname = f.getAbsolutePath();
		newname = newname.substring(0, newname.lastIndexOf(".") + 1) + "pbs";
		f.renameTo(new File(newname));
		log.info("Renamed update file from: " + f.getAbsolutePath() + " to: " + newname);
		
		// Save the updated data file
		save();
		
		// Update the employee name in case this is the
		// first time we've done an update
		parseEmployee();
		
		return i;
		
	}
	
	/** Saves the data to disk */
	public void save() {
		data.saveToFile(new File(DATAFILE_NAME));
		log.info("Saved updated datafile to " + DATAFILE_NAME);
	}
	
	public String getEmployeeName() {
		return ename;
	}
	
	public String getEmployeeForename() {
		return efirstname;
	}
	
	public String getEmployeeSurname() {
		return esurname;
	}
	
	public String getEmployeeCode() {
		return ecode;
	}
	
	public String getPersonalTimeCode() {
		return ptcode;
	}

	/**
	 * Retrieves a list of all projects the employee is
	 * assigned to.
	 * 
	 * @return A list of projects
	 */
	public Project[] getProjects() {
		CSVRow[] c = data.find(0, "EJ");
		if (c.length == 0)
			return new Project[0];
		
		Project[] p = new Project[c.length];
		for (int i = 0; i < c.length; i++) {
			p[i] = new Project();
			p[i].parse(c[i]);
		}
		return p;
	}
	
	/**
	 * Returns available project phases
	 * @return
	 */
	public Phase[] getPhases() {
		CSVRow[] c = data.find(0, "JS");
		if (c.length == 0)
			return new Phase[0];
		
		Phase[] p = new Phase[c.length];
		for (int i = 0; i < c.length; i++) {
			p[i] = new Phase();
			p[i].parse(c[i]);
		}
		return p;
	}
	
	public Task[] getTasks() {
		CSVRow[] c = data.find(0, "JA");
		if (c.length == 0)
			return new Task[0];
		
		Task[] t = new Task[c.length];
		for (int i = 0; i < c.length; i++) {
			t[i] = new Task();
			t[i].parse(c[i]);
		}
		return t;
	}
	
	/**
	 * Returns available expense types
	 * @return
	 */
	public ExpenseType[] getExpenseTypes() {
		CSVRow[] c = data.find(0, "D");
		if (c.length == 0)
			return new ExpenseType[0];
		
		ExpenseType[] e = new ExpenseType[c.length];
		for (int i = 0; i < c.length; i++) {
			e[i] = new ExpenseType();
			e[i].parse(c[i]);
		}
		return e;
	}
	
	public VAT[] getVAT() {
		CSVRow[] c = data.find(0, "V");
		if (c.length == 0)
			return new VAT[0];
		
		VAT[] v = new VAT[c.length];
		for (int i = 0; i < c.length; i++) {
			v[i] = new VAT();
			v[i].parse(c[i]);
		}
		return v;
	}
	
	/**
	 * Return expense records in the data
	 * @return
	 */
	public Expense[] getExpenses() {
		CSVRow[] c = data.find(0, "EXP");
		if (c.length == 0)
			return new Expense[0];
		
		Expense[] e = new Expense[c.length];
		for (int i = 0; i < c.length; i++) {
			e[i] = new Expense();
			e[i].parse(c[i]);
		}
		return e;
	}
	
	/**
	 * Returns all of the expenses applicable for a particular
	 * project code and date.
	 * @param project
	 * @param date
	 * @return
	 */
	public Expense[] getExpensesForProjectAndDate(String project, String date) {
		CSVRow[] e = data.find(0, "EXP");
		CSVRow[] ed = CSVMap.find(e, 3, date);
		CSVRow[] ep = CSVMap.find(ed, 4, project);
		if (ep.length == 0)
			return new Expense[0];
		Expense[] exp = new Expense[ep.length];
		for (int i = 0; i < ep.length; i++) {
			exp[i] = new Expense();
			exp[i].parse(ep[i]);
		}
		return exp;
	}
	
	/**
	 * Return timesheet lines in the data
	 * @return
	 */
	public Timesheet[] getTimesheets() {
		CSVRow[] c = data.find(0, "T");
		if (c.length == 0)
			return new Timesheet[0];
		
		Timesheet[] t = new Timesheet[c.length];
		for (int i = 0; i < c.length; i++) {
			t[i] = new Timesheet();
			t[i].parse(c[i]);
		}
		return t;
	}

	/**
	 * Return timesheet lines in the data between
	 * two dates.
	 */
	public Timesheet[] getTimesheets(Date from, Date to) {
		CSVRow[] c = data.find(0, "T");
		if (c.length == 0)
			return new Timesheet[0];
		
		Vector t = new Vector();

		for (int i = 0; i < c.length; i++) {
			Timesheet s = new Timesheet();
			s.parse(c[i]);
			
			Date d = null;
			try {
				d = new SimpleDateFormat(DATE_FORMAT).parse(s.getDate());
			}
			catch (Exception e) {}
			
			if ( (d.after(from) && d.before(to)) || d.equals(to) || d.equals(from) )
				t.add(s);	
		}
		
		Timesheet[] r = new Timesheet[t.size()];
		for (int i = 0; i < t.size(); i++)
			r[i] = (Timesheet) t.get(i);
		
		return r;
	}
	
	/**
	 * Sorts timesheets into project order.
	 * @param t
	 * @return The sorted list of timesheets
	 */
	public Timesheet[] sortTimesheetsByProject(Timesheet[] t) {
		
		boolean changed = true;
		Timesheet tmp = null;
		while (changed) {
			changed = false;
			for (int i = 0; i < t.length-1; i++) {
				if (t[i].getProject().compareTo(t[i+1].getProject()) > 0) {
					tmp = t[i];
					t[i] = t[i+1];
					t[i+1] = tmp;
					changed = true;
				}
			}
		}
		
		return t;
	}
	
	/**
	 * @param project
	 * @param phase
	 * @return true if the phase is valid for the project
	 */
	public boolean isPhaseInProject(String project, String phase) {
		Phase[] p = getPhasesForProject(project);
		
		// If the project has no phases, then the phase the user
		// selected must be a blank.
		if (p.length == 0)
			if (phase.trim().equals(""))
				return true;
		
		for (int i = 0; i < p.length; i++) {
			if (p[i].getCode().equals(phase))
				return true;
		}
		return false;
	}
	
	/**
	 * @param project
	 * @param task
	 * @return true if the task is valid for the project
	 */
	public boolean isTaskInProject(String project, String task) {
		Task[] t = getTasksForProject(project);
		
		// If the project has no tasks, then the task the user
		// selected must be a blank.
		if (t.length == 0)
			if (task.trim().equals(""))
				return true;
		
		for (int i = 0; i < t.length; i++) {
			if (t[i].getCode().equals(task))
				return true;
		}
		return false;		
	}
	
	/** 
	 * Returns a list of available phases for the given project
	 * @param project
	 * @return
	 */
	public Phase[] getPhasesForProject(String project) {
		Vector v = new Vector();
		Phase[] p = getPhases();
		for (int i = 0; i < p.length; i++) {
			if (p[i].getProject().equals(project))
				v.addElement(p[i]);
		}
		Phase[] o = new Phase[v.size()];
		for (int i = 0; i < v.size(); i++) {
			o[i] = (Phase) v.get(i);
		}
		return o;
	}
	
	/**
	 * Returns a list of available tasks for the given project
	 * @param project
	 * @return
	 */
	public Task[] getTasksForProject(String project) {
		Vector v = new Vector();
		Task[] t = getTasks();
		for (int i = 0; i < t.length; i++) {
			if (t[i].getProject().equals(project))
				v.addElement(t[i]);
		}
		Task[] o = new Task[v.size()];
		for (int i = 0; i < v.size(); i++) {
			o[i] = (Task) v.get(i);
		}
		return o;
	}
	
	
	
	/**
	 * Creates a new expense from the passed parameters, parses
	 * it into a CSVRow and saves the updated datafile.
	 * 
	 * @param expenseType
	 * @param date
	 * @param project
	 * @param quantity
	 * @param amount
	 * @param vatCode
	 * @param phaseCode
	 * @param taskCode
	 * @param notes
	 */
	public void createExpense(String expenseType, String date, String project,
			                  String quantity, String amount, String vatCode,
			                  String phaseCode, String taskCode, String notes) {
		
		Expense e = new Expense();
		e.setExpenseTypeCode(expenseType);
		
		// Look up the description for the type
		CSVRow[] c = data.find(0, "D");
		c = CSVMap.find(c, 1, expenseType);
		e.setExpenseTypeDescription(c[0].getValue(2));
		e.setDate(date);
		e.setProject(project);
		e.setQuantity(quantity);
		e.setAmount(amount);

		c = data.find(0, "V");
		c = CSVMap.find(c, 1, vatCode);
		e.setVatCode(vatCode);
		e.setVatDescription(c[0].getValue(2));
		
		e.setEmployeeCode(getEmployeeCode());
		e.setEmployeeForename(getEmployeeForename());
		e.setEmployeeSurname(getEmployeeSurname());
		
		e.setPhaseCode(phaseCode);
		e.setTaskCode(taskCode);
		e.setNotes(notes);
		e.setSequence(Integer.toString(nextSequence));
		
		data.append(e.toCSVRow());
		data.saveToFile(new File(DATAFILE_NAME));
		log.info("Saved updated data file.");
	}
	
	public void createTimesheet(String date, String project, String hours,
			                    String phaseCode, String taskCode, String notes) {

		Timesheet t = new Timesheet();
		t.setEmployeeCode(getEmployeeCode());
		t.setEmployeeForename(getEmployeeForename());
		t.setEmployeeSurname(getEmployeeSurname());
		t.setDate(date);
		t.setProject(project);
		t.setHours(hours);
		t.setPhaseCode(phaseCode);
		t.setTaskCode(taskCode);
		t.setNotes(notes);
		t.setSequence(Integer.toString(nextSequence));
		
		data.append(t.toCSVRow());
		log.info("Created timesheet row: " + t.toString());
		
		// Save the data
		save();
	}

	/**
	 * Posts time between two dates.
	 * 
	 * Note that we write the CSV output file manually here
	 * because PersonalTime doesn't seem to do "standard" CSV,
	 * and we reproduce faithfully the quirks the Windows client
	 * has.
	 * 
	 * @param from The from date
	 * @param to The to date
	 * @throws IOException if a problem occurred writing to the output file
	 */
	public void postTime(Date from, Date to) throws IOException {
		
		// The output file we're going to write.
		StringBuffer b = new StringBuffer();
		
		b.append("IMPORT2\r\n");
		b.append("W,1\r\n");
		b.append("R, \"" + ptcode + "\"\r\n");
		b.append("\r\n");
		
		Timesheet[] t = getTimesheets();
		
		// Sort the timesheets by project code to make sure
		// the line number algorithm is right
		t = sortTimesheetsByProject(t);
		
		// The "real" PersonalTime UI stamps a line number for
		// each project within a timesheet date range. To emulate
		// this behaviour, every time the project changes during
		// posting time (don't try to post more than a week at
		// once!) we increment the line number to prevent CentralTime
		// getting confused and marking the week to every single
		// project.
		//
		// Eg: In the real client:
		// MYPROJ	PHASE	TASK	7.5	      7.5  7.5  7.5		LINE 1
		// MYPROJ2  PHASE	TASK	     7.5					LINE 2
		// 
		int line = 0;
		
		String lastProject = "";
		for (int i = 0; i < t.length; i++) {
			try {
				Date d = new SimpleDateFormat(DATE_FORMAT).parse(t[i].getDate());
				// Is our date between from and to (or equal to either)?
				if (
						(d.after(from) && d.before(to)) ||
						(d.equals(from)) ||
						(d.equals(to))
					) {
					
					// If the project changes, increment the line number we
					// send to the timesheet server.
					if (!lastProject.equals(t[i].getProject())) {
						line++;
						lastProject = t[i].getProject();
					}
					
					// One append line per column of data plus next separator --
					b.append("T, ");
					b.append(Integer.parseInt(getEmployeeCode()) + ",");
					b.append("\"" + getEmployeeForename() + "\",");
					b.append("\"" + getEmployeeSurname() + "\",");
					b.append(t[i].getDate() + ",");
					b.append("\"" + t[i].getProject() + "\", ");
					b.append(t[i].getHours() + ",");
					b.append("\"" + t[i].getPhaseCode() + "\",");
					b.append("\"" + t[i].getTaskCode() + "\",");
					b.append("\"" + t[i].getNotes() + "\", ");
					b.append(line + ",");
					b.append(t[i].getSequence());
					b.append("\r\n");
					
					// Does this line have any expenses? If so
					// output them now
					Expense[] e = getExpensesForProjectAndDate(t[i].getProject(), t[i].getDate());
					
					// Again, one append per column of data.
					if (e.length > 0) {
						for (int z = 0; z < e.length; z++) {
							b.append("D,");
							b.append(Integer.parseInt(e[z].getExpenseTypeCode()));
							b.append(",\"" + e[z].getExpenseTypeDescription() + "\",");
							b.append(e[z].getDate() + ",");
							b.append("\"" + e[z].getProject() + "\",");
							b.append(" " + e[z].getQuantity() + ",");
							b.append(e[z].getAmount() + ",");
							
							// TODO: Not the right place for VAT - crashes CT here
							// so I need another capture from a real client with expenses/tax
							// on.
							//b.append(Integer.parseInt(e[z].getVatCode()) + ",");
							//b.append("\"" + e[z].getVatDescription() + "\",");
							b.append(",,"); // Leave blank for now
							
							b.append(","); // What is this value? Blank in capture
							b.append(Integer.parseInt(e[z].getEmployeeCode()) + ",");
							b.append("\"" + e[z].getEmployeeForename() + "\",");
							b.append("\"" + e[z].getEmployeeSurname() + "\",");
							b.append("\"" + e[z].getPhaseCode() + "\",");
							b.append("\"" + e[z].getTaskCode() + "\",");
							b.append("\"" + e[z].getNotes() + "\",");
							b.append(e[z].getSequence() + ",");
							b.append("0,"); // What is this value? Always 0 in capture
							
							// Don't know what this value is - seen 4 on non-mileage
							// and 8 on mileage
							if (e[z].getAmount().equals(""))
								b.append("8,");
							else
								b.append("4,");
							
							// Last four values are empty in my sample captures
							b.append(",");
							b.append(",");
							b.append(",");
							b.append(",");
							b.append("\r\n");
						}
					}
						
					e = null;
				}
			}
			catch (ParseException e) {
				// Shouldn't happen - we've checked the
				// dates parse already
			}
		}
		
		// Create the output file
		File f = new File(sharedDirectory +  File.separator + group + ".ptr");
		FileOutputStream out = new FileOutputStream(f);
		out.write(b.toString().getBytes());
		out.flush();
		out.close();
				
		b = null;
		
		// Increment the sequence number
		incrementSequence();
	}
	
	/**
	 * Increments the sequence number for the next
	 * timesheet to be posted and saves the updated datafile.
	 */
	public void incrementSequence() {
		// Increment the next sequence number
		nextSequence++;
		log.info("Incremented next timesheet sequence to " + nextSequence);
		CSVRow[] c = data.find(0, "SEQ");
		c[0].putValue(1, Integer.toString(nextSequence));
		save();
	}
	
}
