package net.sf.freetime;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Map;
import java.util.Vector;

import net.sf.freetime.dto.Expense;
import net.sf.freetime.dto.ExpenseType;
import net.sf.freetime.dto.Phase;
import net.sf.freetime.dto.Project;
import net.sf.freetime.dto.Task;
import net.sf.freetime.dto.Timesheet;
import net.sf.freetime.dto.VAT;
import net.sf.ooweb.HTTP;
import net.sf.ooweb.util.Logger;

/**
 * Main class implementing FreeTime UI.
 * 
 * @author Robin Rawson-Tetley
 */
public class FreeTime {

	/** The OOWeb logger */
	private Logger log = null;
	
	/** Central data containing all records */
	private DataController data = null;
	
	/** The format all dates should be in for validation purposes */
	private final static String DATE_FORMAT = "dd/MM/yyyy";
	
	public FreeTime(Logger log) {
		this.log = log;
		this.data = new DataController(log);
	}
	
	// Pages
	
	/**
	 * Lists what we can do and projects we're
	 * assigned to.
	 */
	public String index() {
		StringBuffer b = new StringBuffer();
		b.append("<html><head><title>FreeTime!</title>" +
			   "</head><body>" +
			   "<h2>FreeTime</h2>" +
			   "<p>A free, minimal, cross-platform, GPLed client for the <b>unbelievably bad Timewatch " +
			   " PersonalTime</b> product. If your company uses this, I feel your pain.</p>" +
			   "<ul>" +
			   "<li><a href=\"update\">Check for data updates</a></li>" +
			   "<li><a href=\"add\">Add/View timesheet data</a></li>" +
			   "<li><a href=\"post\">Post timesheet data</a></li>" +
			   "</ul>" +
			   "<hr />");
		
        // Find out any projects we're assigned to
		Project[] p = data.getProjects();
		if (p.length > 0) {
			b.append("<p>" + data.getEmployeeName() + " (E" + data.getEmployeeCode() + ") is assigned to:</p>");
			b.append("<ul>");
			for (int i = 0; i < p.length; i++) {
				b.append("<li>");
				b.append(p[i].getCode() + " - " + p[i].getDescription());
				b.append("</li>");
			}
			b.append("</ul>");
		}
		
		b.append("</body></html>");
		return b.toString();
	}
	
	/**
	 * Searches for updates to the timesheet data and
	 * lets the user know how many we found.	 
	 */
	public String update() {
	
		int no = data.updateFromShare();
		
		if (no == 0) {
			return "<html><body><p>No updates were found.</p>" +
			       "<p><a href=\"/\">back</a></p></body></html>";
		}
		else {
			return "<p><b>Found " + no + " updates.</b></p>" +
				 "<p><a href=\"/\">back</a></p></body></html>";
		}
	}
	
	/**
	 * Allows entering of expenses for a given
	 * day/project
	 */
	public String expenses(Map m) {
		
		StringBuffer b = new StringBuffer();
		b.append("<html><head><title>Expenses</title></head><body>");
		b.append("<h2>Expenses</h2>");
		b.append("<form action=\"expensesAction\" method=\"post\"><table><tr>");
		b.append("<td><b>Description</b></td>");
		b.append("<td><b>Quantity</b></td>");
		b.append("<td><b>Amount</b></td>");
		b.append("<td><b>VAT</b></td>");
		b.append("<td><b>Notes</b></td>");
		b.append("</tr>");
		
		// Find any existing records and output them -
		// note, they are "D" records when posted, but this
		// clashes in our datafile with Expense types (which are
		// also D records - stupid Timewatch)
		Expense[] e = data.getExpenses();
		for (int i = 0; i < e.length; i++) {
			
			// Only include ones which match the
			// date and project code
			if ( e[i].getDate().equals(m.get("date").toString())
				 && e[i].getProject().equals(m.get("project").toString()) ) {
				
				b.append("<tr>");
				b.append("<td>" + e[i].getExpenseTypeDescription() + "</td>");
				b.append("<td>" + e[i].getQuantity() + "</td>");
				b.append("<td>" + e[i].getAmount() + "</td>");
				b.append("<td>" + e[i].getVatCode() + "</td>");
				b.append("<td>" + e[i].getNotes() + "</td>");
				b.append("</tr>");				
			}
		}
		
		// Output the form to enter a new expense
		b.append("<tr>");
		b.append("<td>");
		b.append("<select name=\"type\">");
		ExpenseType[] ex = data.getExpenseTypes();
		for (int i = 0; i < ex.length; i++) {
			b.append("<option value=\"" + ex[i].getCode() + "\">");
			b.append(ex[i].getDescription());
			b.append("</option>");
		}
		b.append("</select></td>");
		
		b.append("<td><input name=\"quantity\"/></td>");
		b.append("<td><input name=\"amount\"/></td>");
		
		b.append("<td><select name=\"vat\">");
		VAT[] v = data.getVAT();
		for (int i = 0; i < v.length; i++) {
			b.append("<option value=\"" + v[i].getCode() + "\">");
			b.append(v[i].getDescription());
			b.append("</option>");
		}
		b.append("</select></td>");
		b.append("<td><textarea name=\"notes\"></textarea></td>");
		b.append("</tr>");
		b.append("</table><p>");
		// Hidden fields containing additional form data
		b.append("<input type=\"hidden\" name=\"date\" value=\"" + m.get("date") + "\"/>");
		b.append("<input type=\"hidden\" name=\"project\" value=\"" + m.get("project") + "\"/>");
		b.append("<input type=\"hidden\" name=\"phase\" value=\"" + m.get("phase") + "\"/>");
		b.append("<input type=\"hidden\" name=\"task\" value=\"" + m.get("task") + "\"/>");
		b.append("<input type=\"submit\" value=\"Submit\"/>");
		b.append("</p></form>");
		b.append("<p><a href=\"/\">back</a></p>");
		b.append("</body></html>");
		
		return b.toString();
	}
	
	/**
	 * Adds an expense record
	 * 
	 * @param m
	 * @return
	 */
	public String expensesAction(Map m) {
		
		data.createExpense(m.get("type").toString(),
				           m.get("date").toString(),
				           m.get("project").toString(),
				           m.get("quantity").toString(),
				           m.get("amount").toString(),
				           m.get("vat").toString(),
				           m.get("phase").toString(),
				           m.get("task").toString(),
				           m.get("notes").toString());
		
		return HTTP.redirect("expenses?date=" + m.get("date").toString() +
					"&project=" + m.get("project").toString() + "&phase=" +
					m.get("phase").toString() + "&task=" + m.get("task").toString());
		
	}
	
	/**
	 * Form where a timesheet can be added.
	 * @param The start of the week or null to use today
	 * @return
	 */
	public String add() {
		
		StringBuffer b = new StringBuffer();
		b.append("<html><head><title>Timesheet Data</title></head><body>");
		b.append("<h2>Timesheet Data</h2>");
			
		// Table header
		b.append("<form method=\"post\" action=\"addAction\">");
		b.append("<table border=1>");
		b.append("<tr>");
		b.append("<td><b>Date</b></td>");
		b.append("<td><b>Project/Phase/Task</b></td>");
		b.append("<td><b>Hours</b></td>");
		b.append("<td><b>Notes</b></td>");
		b.append("<td></td>");
		b.append("</tr>");
		
		// Output existing timesheet data (T records)
		Timesheet[] t = data.getTimesheets();
		if (t.length > 0) {
			for (int i = 0; i < t.length; i++) {
				b.append("<tr>");
				b.append("<td>" + t[i].getDate() + "</td>");
				b.append("<td>" + t[i].getProject() + ", " + 
					t[i].getPhaseCode() + ", " + 
					t[i].getTaskCode() + "</td>");
				b.append("<td>" + t[i].getHours() + "</td>");
				b.append("<td>" + t[i].getNotes() + "</td>");
				b.append("<td><a href=\"expenses?date=" + t[i].getDate() +
					"&project=" + t[i].getProject() + "&phase=" +
					t[i].getPhaseCode() + "&task=" + t[i].getTaskCode() + "\">" +
					"Exp</a></td>");
				b.append("</tr>");
			}
		}

		// Now output the input fields
		b.append("<tr>");
		b.append("<td><input name=\"date\" size=10 " +
				"value=\"" + 
				new SimpleDateFormat(DATE_FORMAT).format(new Date()) + 
				"\" />");
		
		// List of available projects
		b.append("<td>");
		b.append("<select name=\"project\">");
		Project[] p = data.getProjects();
		String selected = "";
		String lastproject = (String) HTTP.getSession().get("project");
		if (lastproject == null) lastproject = "";
		for (int i = 0; i < p.length; i++) {
			if (p[i].getCode().equals(lastproject)) {
				selected = " selected";
			}
			else {
				selected = "";
			}
			b.append("<option" + selected + " value=\"" + p[i].getCode() + 
					"\">" + p[i].getCode() + " - " + p[i].getDescription() + "</option>");
		}
		b.append("</select>");
		
		// List of available phases
		b.append("<select name=\"phase\"><option value=\"\"></option>");
		Phase[] ph = data.getPhases();
		Vector used = new Vector(); // Used to squash dupes
		selected = "";
		String lastphase = (String) HTTP.getSession().get("phase");
		if (lastphase == null) lastphase = "";
		for (int i = 0; i < ph.length; i++) {
			if (!used.contains(ph[i].getCode())) {
				used.addElement(ph[i].getCode());
				if (ph[i].getCode().equals(lastphase)) {
					selected = " selected";
				}
				else {
					selected = "";
				}
				b.append("<option" + selected + " value=\"" + ph[i].getCode() + 
						"\">" + ph[i].getCode() + " - " + ph[i].getDescription() + "</option>");
			}
		}
		b.append("</select>");
		
		// List of available tasks
		b.append("<select name=\"task\"><option value=\"\"></option>");
		Task[] ta = data.getTasks();
		used.clear();
		selected = "";
		String lasttask = (String) HTTP.getSession().get("task");
		if (lasttask == null) lasttask = "";
		for (int i = 0; i < ta.length; i++) {
			if (!used.contains(ta[i].getCode())) {
				used.addElement(ta[i].getCode());
				if (ta[i].getCode().equals(lasttask)) {
					selected = " selected";
				}
				else {
					selected = "";
				}
				b.append("<option" + selected + " value=\"" + ta[i].getCode() + 
						"\">" + ta[i].getCode() + " - " + ta[i].getDescription() + "</option>");
			}
		}
		b.append("</select></td>");
		
		// Hours/Notes
		String hours = (String) HTTP.getSession().get("hours");
		if (hours == null) hours = "0";
		b.append("<td><input name=\"hours\" value=\"" + hours + "\" size=4/></td>");
		
		String notes = (String) HTTP.getSession().get("notes");
		if (notes == null) notes = "";
		b.append("<td><textarea name=\"notes\">" + notes + "</textarea></td>");
		
		b.append("</table>");
		b.append("<input type=\"submit\" value=\"Submit\"/>");
		b.append("</form>");
		
		b.append("<p><a href=\"/\">back</a></p>");
		b.append("</body></html>");
		return b.toString();
	}
	
	/** 
	 *  Validate and add the time record to the main
	 *  data and save.
	 * @param m The map of form data
	 * @return
	 */
	public String addAction(Map m) { 
		
		// Validate the date
		try {
			new SimpleDateFormat(DATE_FORMAT).parse(m.get("date").toString());
		}
		catch (ParseException e) {
			return "<html><body><p>The date you entered was invalid.</p></body></html>";
		}

		// Verify that the phase and task are allowed for the
		// project selected to prevent later barfs.
		
		if (!data.isPhaseInProject(m.get("project").toString(), m.get("phase").toString())) {
			StringBuffer sb = new StringBuffer("<html><body><h2>Error</h2>");
			sb.append("<p>An error occurred, the phase you selected is not " + 
					  "valid for the project.</p><p>Valid phases for ");
			sb.append(m.get("project").toString());
			sb.append(" are:</p><ul>");
			Phase[] ph = data.getPhasesForProject(m.get("project").toString());
			for (int i = 0; i < ph.length; i++) {
				sb.append("<li>" + ph[i].getCode() + " - " + ph[i].getDescription() + "</li>");
			}
			sb.append("</ul></body></html>");
			return sb.toString();
		}
		
		if (!data.isTaskInProject(m.get("project").toString(), m.get("task").toString())) {
			StringBuffer sb = new StringBuffer("<html><body><h2>Error</h2>");
			sb.append("<p>An error occurred, the task you selected is not " + 
					  "valid for the project.</p><p>Valid tasks for ");
			sb.append(m.get("project").toString());
			sb.append(" are:</p><ul>");
			Task[] ta = data.getTasksForProject(m.get("project").toString());
			for (int i = 0; i < ta.length; i++) {
				sb.append("<li>" + ta[i].getCode() + " - " + ta[i].getDescription() + "</li>");
			}
			sb.append("</ul></body></html>");
			return sb.toString();
		}
		
		data.createTimesheet(m.get("date").toString(),
				             m.get("project").toString(),
				             m.get("hours").toString(),
				             m.get("phase").toString(),
				             m.get("task").toString(),
				             m.get("notes").toString());
		
		// Cache the last selections in the user's session
		// so we can default them next time
		HTTP.getSession().put("project", m.get("project"));
		HTTP.getSession().put("phase", m.get("phase"));
		HTTP.getSession().put("task", m.get("task"));
		HTTP.getSession().put("hours", m.get("hours"));
		HTTP.getSession().put("notes", m.get("notes"));

		// Go back to the add page so the user sees it
		return HTTP.redirect("add");
		
	}
	
	/**
	 * Displays a form allowing the user to post time between
	 * two dates. No checking is done for already posted time
	 * and you can post the same time twice so be careful.
	 * 
	 * @return
	 */
	public String post() {
		
		// Make sure we actually have time
		Timesheet[] t = data.getTimesheets();
		if (t.length == 0) {
			return "<html><body><p>Enter some time first!</p><p><a href=\"/\">back</a></p></body></html>";
		}
		
		// Calculate the Monday at the beginning of this
		// week, and the Sunday at the end of the week.
		Calendar c = Calendar.getInstance();
		// Count backwards to Monday
		while (c.get(Calendar.DAY_OF_WEEK) != Calendar.MONDAY) {
			c.set(Calendar.DAY_OF_YEAR, c.get(Calendar.DAY_OF_YEAR) - 1);
		}
		Date fd = c.getTime();
		
		// Go forwards to Friday
		while (c.get(Calendar.DAY_OF_WEEK) != Calendar.FRIDAY) {
			c.set(Calendar.DAY_OF_YEAR, c.get(Calendar.DAY_OF_YEAR) + 1);
		}
		Date td = c.getTime();
		
		// Format the dates for display on the screen
		String from = new SimpleDateFormat(DATE_FORMAT).format(fd);
		String to = new SimpleDateFormat(DATE_FORMAT).format(td);
		
		StringBuffer b = new StringBuffer();
		b.append("<html><head><title>Post Time</title><body>");
		b.append("<h2>Post Time</h2>");
		b.append("<p><b>Note: Due to conversion and numbering " +
		         " problems you should not attempt to post more " +
		         "than one week of time in one go.</b></p>");
		b.append("<form method=\"post\" action=\"postAction\">");
		b.append("<p>Post time from: <input name=\"from\" value=\"" + from + "\"/></p>");
		b.append("<p>Post time to: <input name=\"to\" value=\"" + to + "\"/></p>");
		b.append("<p><input type=\"submit\" value=\"Submit\"/>");
		b.append("</body></html>");
		
		return b.toString();
		
	}
	
	/**
	 * Posts the time
	 * @param m
	 * @return
	 */
	public String postAction(Map m) {
		
		// Loop through all of our time records, work out whether
		// the dates match ok and build the output
		Date from = null;
		Date to = null;
		
		try {
			from = new SimpleDateFormat(DATE_FORMAT).parse(m.get("from").toString());
			to = new SimpleDateFormat(DATE_FORMAT).parse(m.get("to").toString());
		}
		catch (ParseException e) {
			return "One of the dates you entered was invalid.";
		}
		
		try {
			data.postTime(from, to);
		}
		catch (Exception e) {
			log.error(e);
			return "Failed to write the communication file. See the log for details.";
		}
		
		return "<html><body><p>Time successfully posted.</p></p><a href=\"/\">back</a></p></body></html>";
		
	}

}
